package com.mhsoftware.jonnie;
/*
    MH Software, Inc.
    5023 W 120th Ave, #311
    Broomfield, CO 80020

    Voice: 303 438 9585
    Web: http://www.mhsoftware.com/

    Copyright 2008, MH Software, Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.text.NumberFormat;
import java.util.Random;

/** 
 * This class implements a file creation test. It creates the
 * specified number of files, writes the specified amount of
 * data to them, and then reads the files and verifies the data
 * was written correctly.
 * 
 * @author George Sexton
 * 
 */
public class FileCreationTest extends JonnieTest {

    private long lSeed=Long.MIN_VALUE;

    private void corruptFile(String sFileNameRoot,final int fileCount,final int iSize) throws IOException {
        final int iCorruptFile=(int)(Math.random()*fileCount);
        final int iOffset=(int)(Math.random()*iSize);
        final String sFile=sFileNameRoot+iCorruptFile+".tmp";

        RandomAccessFile raf=new RandomAccessFile(new File(sFile),"rw");
        raf.seek(iOffset);
        byte b=raf.readByte();
        System.err.println("Corrupting File ["+sFile+"] at offset "+iOffset+" changing "+Byte.toString(b)+" to "+Byte.toString((byte)(b^0x7f)));
        raf.seek(iOffset);
        raf.writeByte(b^0x7f);
        raf.close();
    }

    /**
     * Run the actual test. Implementations must call preRun(),
     * and postRun() at the appropriate points in their method.
     * <BR><BR>
     * Additionally, implementations should check the
     * isPendingInterrupt() method to see if they should terminate
     * processing because of an interrupt.
     */
    public void run() {

        Throwable tError=null;
        final int iFiles=(int)TestUtil.getPowerOf2Value(getProperty("FileCount","1k"));
        final int iSize=(int)TestUtil.getPowerOf2Value(getProperty("FileSize","1k"));
        final boolean verifyData="true".equalsIgnoreCase(getProperty("VerifyData","true"));
        final String sFileNameRoot=getProperty("TestPath")+File.separator+Thread.currentThread().getName()+"-file";
        lSeed=System.currentTimeMillis();
        Random prng=new Random(lSeed);
        
        int i;
        FileChannel chan;
        ByteBuffer bufWrite=ByteBuffer.allocateDirect(iSize);

        preRun();
        try {

            for (i=0; i < iFiles && !this.isPendingInterrupt(); i++) {
                BlockDataGenerator.fillBufferRandom(bufWrite,prng);
                FileOutputStream fos=new FileOutputStream(sFileNameRoot+i+".tmp");
                chan=fos.getChannel();
                chan.write(bufWrite);
                fos.close();
                bufWrite.rewind();
            }


            if (testCorruption()) {
                corruptFile(sFileNameRoot,iFiles,iSize);
            }

            if (!this.isPendingInterrupt()) {
                prng=new Random(lSeed);
                ByteBuffer bufRead=ByteBuffer.allocateDirect(iSize);
                for (i=0; i <iFiles && !this.isPendingInterrupt(); i++) {
                    File f=new File(sFileNameRoot+i+".tmp");
                    FileInputStream fis=new FileInputStream(f);
                    chan=fis.getChannel();
                    bufRead.clear();
                    chan.read(bufRead);
                    if (verifyData) {
                    	BlockDataGenerator.fillBufferRandom(bufWrite,prng);
                        bufRead.flip();
                        if (!bufWrite.equals(bufRead)) {
                            DataVerificationError dve=new DataVerificationError("Failure verifying read data! File["+f.getName()+"]");
                            this.reportException(dve);
                            throw dve;
                        }
                    }
                    fis.close();
                    chan.close();
                    if (!f.delete()) {
                        throw new IOException("Error deleting file "+f);
                    }
                    
                }
            }
            double dResult=((double)iFiles/((double)this.getElapsedTimeMillis()/1000.0));
            NumberFormat nf=NumberFormat.getInstance();
            TestResult tr=new TestResult(
                                        "Files Creation",
                                        nf.format(dResult),
                                        "Files/Second");
            this.addResult(tr);

        } catch (IOException ioe) {
            tError=ioe;
            reportException(ioe);
        }
        if (tError!=null) {
            DataVerificationError dve=new DataVerificationError("FileCreationTest.run()",tError);
            this.reportException(dve);
            throw dve;
        }
        postRun();
    }
}
