package com.mhsoftware.jonnie;
/*
    MH Software, Inc.
    5023 W 120th Ave, #311
    Broomfield, CO 80020

    Voice: 303 438 9585
    Web: http://www.mhsoftware.com/

    Copyright 2008, MH Software, Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

abstract class JonnieTest implements Cloneable {

    private List<TestResult> lstResults=null;
    private List<TestEventListener> lstListeners=null;
    private Properties props=null;
    private boolean pendingInterrupt;
    private boolean running=false;
    private boolean dirtyShutdown=false;
    private long lStart,lEnd;


    /** 
     * Return a shallow copy clone of this test object.
     * 
     * 
     * @return Object
     */
    public Object clone() throws CloneNotSupportedException {
        JonnieTest jtCopy=(JonnieTest)super.clone();
        jtCopy.clearResults();
        return jtCopy;
    }

    /**
     * Add an event listener for this test. 
     * @param lstnr The listener to add.  
     */
    public void addEventListener(TestEventListener lstnr){
        if (lstListeners==null) {
            lstListeners=new ArrayList<TestEventListener>();
        }
        lstListeners.add(lstnr);
    }

    /**
     * Add a property for this test. Properties are used to 
     * configure run-time parameters for the test. 
     */
    public void addProperty(String sName, String sValue){
        if (props==null) {
            props=new Properties();
        }
        props.setProperty(sName,sValue);
    }

    public void addProperties(Properties propIn){
        if (props==null) {
            props=new Properties();
        }
        props.putAll(propIn);
    }

    public String getProperty(String sName,String sDefault){
        if (props==null) {
            return sDefault;
        } else {
            return props.getProperty(sName,sDefault);
        }
    }
    /**
     * Get a property for this test. If the property is not found, 
     * null is returned. 
     */
    public String getProperty(String sName){
        if (props==null) {
            return null;
        } else {
            return props.getProperty(sName);
        }
    }

    public Properties getProperties(){
        if (props==null) {
            props=new Properties();
        }
        return props;
    }

    public boolean isRunning(){
        return running;
    }

    public void interrupt(){
        pendingInterrupt=true;
    }

    public boolean isPendingInterrupt(){
        return pendingInterrupt;
    }

    protected void addResult(TestResult res){
        if (lstResults==null) {
            lstResults=new ArrayList<TestResult>();
        }
        lstResults.add(res);
    }

    public void clearResults(){
        lstResults=new ArrayList<TestResult>();
    }

    /** 
     * Return true if the configuration flag to test corruption of 
     * data is set. 
     * 
     * 
     * @return boolean
     */
    public boolean testCorruption(){
        return "true".equalsIgnoreCase(getProperty("TestCorruption")) || "yes".equalsIgnoreCase(getProperty("TestCorruption"));
    }


    public TestResult[] getResults(){
        TestResult[] aRes=new TestResult[lstResults.size()];
        aRes=lstResults.toArray(aRes);
        return aRes;
    }

    /** 
     * Get the elapsed time in milli-seconds. If the test is still 
     * running, the current elapsed time is returned. If the test 
     * has finished, returns the time elapsed during the test. 
     * 
     * @return long
     */
    public long getElapsedTimeMillis(){
        if (this.isRunning()) {
            return System.currentTimeMillis()-lStart;
        } else {
            return lEnd-lStart;
        }
    }

    protected void reportException(Exception e){
        dirtyShutdown=true;
        if (lstListeners==null) {
            System.out.println(e);
        } else {
            for (TestEventListener tel : lstListeners) {
                tel.testExceptionEvent(e,this);
            }
        }

    }

    /** 
     * Child classes run() method should call this before the start
     * of processing.
     * 
     */
    protected void preRun(){
        pendingInterrupt=false;
        dirtyShutdown=false;
        running=true;
        clearResults();
        if (lstListeners!=null) {
            for (TestEventListener tel : lstListeners) {
                tel.testBegin(this);
            }
        }
        lStart=System.currentTimeMillis();
        lEnd=lStart;
    }


    /** 
     * Child classes run() method should calls this after the end of
     * processing.
     * 
     */
    protected void postRun(){
        lEnd=System.currentTimeMillis();
        running=false;
        if (lstListeners!=null) {
            for (TestEventListener tel : lstListeners) {
                tel.testEnd(this);
                if (!this.isPendingInterrupt() && !dirtyShutdown) {
                    tel.testComplete(this);
                }
            }
        }
        pendingInterrupt=false;
    }

    /** 
     * Run the actual test. Implementations must call preRun(),
     * and postRun() at the appropriate points in their method.
     * <BR><BR> 
     * Additionally, implementations should check the 
     * isPendingInterrupt() method to see if they should terminate 
     * processing because of an interrupt. 
     */
    public abstract void run();

}
