package com.mhsoftware.jonnie;
/*
    MH Software, Inc.
    5023 W 120th Ave, #311
    Broomfield, CO 80020

    Voice: 303 438 9585
    Web: http://www.mhsoftware.com/

    Copyright 2008, MH Software, Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.text.NumberFormat;

import java.util.Arrays;
import java.util.Random;

/** 
 * This class performs a Random IO Test. It first creates a file
 * of the specified size, and initializes it with random data.
 * <BR><BR>
 * Next, it does a random write test to the file. The number of
 * writes is the BlockCount configuration value. The positions 
 * to write to are randomly generated from a java.util.random. 
 * <BR><BR>
 * After the write test completes, a read test is conducted,
 * again making BlockCount reads, from locations
 * randomly generated. The goal is to force the disk sub-system
 * to perform seeks across the file. 
 *  <BR><BR>
 * Once the random IO tests are completed, the underlying areas 
 * of the file that were not written are verified as having 
 * their original random data. 
 * <BR><BR>
 * To be a valid exercise, the size of the file should be at
 * LEAST 4-8 times larger than the physical ram of the test
 * machine.
 *  
 * @author George Sexton 
 */
public class RandomIOTest extends JonnieTest {

    private Random rand;
    private long lSeed;
    private  long[] aBlocks;

    private void initFile(File f,ByteBuffer bufData,final long lBlockCount,Random prng) throws IOException {
        FileOutputStream fos=new FileOutputStream(f);
        FileChannel fc=fos.getChannel();
        for (long lBlock=0;  !isPendingInterrupt() &&  (lBlock < lBlockCount); lBlock++) {
            BlockDataGenerator.fillBufferRandom(bufData,prng);
            fc.write(bufData);
            bufData.rewind();
        }
        fos.close();
    }

    private void verifyBlocksNotWrittenRandomly(File f,ByteBuffer bufData,final int iBlockSize,final long lBlockCount,Random prng) throws IOException {
        Arrays.sort(aBlocks);
        FileInputStream fis=new FileInputStream(f);
        FileChannel fc=fis.getChannel();
        ByteBuffer bufRead=ByteBuffer.allocateDirect(iBlockSize);
        for (long lBlock=0; !isPendingInterrupt() && (lBlock<lBlockCount); lBlock++) {
            BlockDataGenerator.fillBufferRandom(bufData,prng);
            fc.read(bufRead);
            bufRead.flip();
            if (Arrays.binarySearch(aBlocks,lBlock)<0) {
                /* This block was not written to as part of the random write testing, verify it. */
                if (!bufData.equals(bufRead)) {
                    DataVerificationError dve=new DataVerificationError(
                                                                       "Data Verify Failure. Random IO NonWritten Area Read Test. File Name: "+f.getName()+" Block="+lBlock+" Block Size="+iBlockSize
                                                                       );
                    this.reportException(dve);
                    throw dve;
                }
            }
        }

        fis.close();
    }

    /** 
     * Generate a list of random blocks for Read/Write
     * 
     * @param iBlockCount The number of blocks for read/write 
     * @param lFileBlocks The number of blocks in the file. 
     * 
     * @return long[] Array of random blocks.
     */
    private long[] generateBlockList(Random prng,
                                     final int iBlockCount,
                                     final long lFileBlocks){
        long[] aResult=new long[iBlockCount];
        long lBlock;
        int iCurrentPos=0;
        nextValue:
        while (iCurrentPos<iBlockCount) {
            lBlock=prng.nextLong();
            if (lBlock<0) {
                continue;
            }
            lBlock%=lFileBlocks;
            for (int iSearch=0; iSearch<iCurrentPos; iSearch++) {
                if (aResult[iSearch]==lBlock) {
                    continue nextValue;
                }
            }
            aResult[iCurrentPos++]=lBlock;
        }
        return aResult;
    }

    private long performWriteTests(File f,ByteBuffer bufData, final long lBlockCount) throws IOException {

        final int iBlockWrites=(int)TestUtil.getPowerOf2Value(getProperty("BlockCount","32k"));

        aBlocks=generateBlockList(new Random(lSeed),iBlockWrites,lBlockCount);
        long lStart=getElapsedTimeMillis();
        String sMode="rw";
        if ("true".equalsIgnoreCase(getProperty("SyncOnWrite","false"))) {
            sMode+="d";
        }
        RandomAccessFile raf=new RandomAccessFile(f,sMode);
        FileChannel fc=raf.getChannel();
        long lPosition;
        final int bufSize=bufData.capacity();

        for (int iBlock=0;  !isPendingInterrupt() &&  (iBlock<iBlockWrites); iBlock++) {
            lPosition=aBlocks[iBlock]*bufSize;
            fc.position(lPosition);
            BlockDataGenerator.fillBufferRandom(bufData,rand);
            fc.write(bufData);
            bufData.rewind();
        }
        raf.close();
        return getElapsedTimeMillis()-lStart;
    }

    private long performReadTests(File f,ByteBuffer bufCompare,final boolean bVerify) throws IOException {
        long lStart=getElapsedTimeMillis();
        RandomAccessFile raf=new RandomAccessFile(f,"r");
        FileChannel fc=raf.getChannel();
        long lPosition;
        final int bufSize=bufCompare.capacity();
        ByteBuffer bufRead=ByteBuffer.allocateDirect(bufSize);
        for (int iBlock=0,iLimit=aBlocks.length;  !isPendingInterrupt() &&  (iBlock<iLimit); iBlock++) {

            lPosition=aBlocks[iBlock]*(long)bufSize;

            fc.position(lPosition);
            fc.read(bufRead);
            BlockDataGenerator.fillBufferRandom(bufCompare,rand);
            if (bVerify) {
                bufRead.flip();
                if (!bufCompare.equals(bufRead)) {
                    DataVerificationError dve=new DataVerificationError(
                                                                       "Data Verify Failure. RandomIO Read Test File Name: "+f.getName()+" Position="+lPosition+" Block Size="+bufSize
                                                                       );
                    this.reportException(dve);
                    throw dve;
                }
            }
            bufRead.clear();
        }
        raf.close();
        return getElapsedTimeMillis()-lStart;
    }

    private void corruptFile(File f,final long lFileSize) throws IOException{
        long lCorruptPos;
        while ((lCorruptPos=rand.nextLong())<0) {
            ;   // Wait for a posive offset to corrupt...
        }
        lCorruptPos%=lFileSize;
        System.err.println("Corrupting File "+f+" at position "+lCorruptPos);
        RandomAccessFile raf=new RandomAccessFile(f,"rw");
        raf.seek(lCorruptPos);
        byte b=raf.readByte();
        raf.seek(lCorruptPos);
        raf.writeByte(b^0x7f);
        raf.close();
    }


    /**
     * Run the actual test. Implementations must call preRun(),
     * and postRun() at the appropriate points in their method.
     * <BR><BR>
     * Additionally, implementations should check the
     * isPendingInterrupt() method to see if they should terminate
     * processing because of an interrupt.
     */
    public void run() {
        Throwable tError=null;
        lSeed=System.currentTimeMillis();
        File f=new File(getProperty("TestPath",".")+File.separator+Thread.currentThread().getName()+"-RandomReadWriteTest.tmp");
        final long lFileSize=TestUtil.getPowerOf2Value(getProperty("FileSize","1G"));
        final int iBlockSize=(int)TestUtil.getPowerOf2Value(getProperty("BlockSize","32k"));
        final long lBlockCount=lFileSize/iBlockSize;
        final long lReadWrites=TestUtil.getPowerOf2Value(getProperty("BlockCount","32k"));
        final boolean bVerify="true".equalsIgnoreCase(getProperty("VerifyRead","true"));
        NumberFormat nf=NumberFormat.getInstance();       

        ByteBuffer bufWrite=ByteBuffer.allocateDirect(iBlockSize);
        rand=new java.util.Random(lSeed);

        try {
            preRun();
            initFile(f,bufWrite,lBlockCount,rand);

            rand=new java.util.Random(lSeed);
            long lWrite=performWriteTests(f,bufWrite,lBlockCount);

            if (testCorruption()) {
                corruptFile(f,lFileSize);
            }

            rand=new java.util.Random(lSeed);
            long lRead=performReadTests(f,bufWrite,bVerify);


            final double dWrite=(double)lWrite/1000.0;
            final double dRead=(double)lRead/1000.0;
            double dValue=(double)lReadWrites/dRead;
            TestResult tr=new TestResult("Random Access Reads",nf.format(dValue),"Reads/Second");
            addResult(tr);
            double dBytesSecond=dValue*iBlockSize;
            tr=new TestResult("Random Access Read Throughput",nf.format(dBytesSecond/(double)(1024*1024)),"MB/Sec");
            addResult(tr);
            dValue=(double)lReadWrites/dWrite;
            tr=new TestResult("Random Access Writes",nf.format(dValue),"Writes/Second");
            addResult(tr);
            dBytesSecond=dValue*iBlockSize;
            tr=new TestResult("Random Access Write Throughput",nf.format(dBytesSecond/(double)(1024*1024)),"MB/Sec");
            addResult(tr);

            if (bVerify) {
                verifyBlocksNotWrittenRandomly(f,bufWrite,iBlockSize,lBlockCount,new Random(lSeed));
            }
            if (!f.delete()) {
                throw new IOException("Error deleting file "+f);
            }
        } catch (IOException ioe) {
            tError=ioe;
            reportException(ioe);
        }
        if (tError!=null) {
            DataVerificationError dve=new DataVerificationError("RandomIOTest.run()",tError);
            this.reportException(dve);
            throw dve;
        }
        postRun();
    }
}
