package com.mhsoftware.jonnie;
/*
    MH Software, Inc.
    5023 W 120th Ave, #311
    Broomfield, CO 80020

    Voice: 303 438 9585
    Web: http://www.mhsoftware.com/

    Copyright 2008, MH Software, Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.text.NumberFormat;
import java.util.Random;
/** 
 * This class implements a sequential Read/Write test. The
 * default file size is 1GB 
 * <BR><BR> 
 *  The file size argument should be AT LEAST 4-8 times larger
 *  than physical memory. If it's not, then what will get
 *  reported is the cached file read time which will be
 *  ridiculously large.
 */
public class SequentialIOTest extends JonnieTest {

    private long lSeed;

    private long performReadTest(File f,ByteBuffer bufCompare,final long lReads, final boolean bVerify,Random prng) 
    throws IOException {
        ByteBuffer bufRead=ByteBuffer.allocateDirect(bufCompare.capacity());
        long lStart=getElapsedTimeMillis();
        FileInputStream fis=new FileInputStream(f);
        FileChannel fc=fis.getChannel();
        for (long lBlock=0;  !isPendingInterrupt() &&  (lBlock < lReads); lBlock++) {
            fc.read(bufRead);
            if (bVerify) {
                bufRead.flip();
                if (prng!=null) {
                    BlockDataGenerator.fillBufferRandom(bufCompare,prng);
                }

                if (!bufCompare.equals(bufRead)) {
                    DataVerificationError dve=new DataVerificationError("Failure verifying read data! File="+f.getName()+" Block="+lBlock+" BlockSize="+bufRead.capacity());
                    this.reportException(dve);
                    throw dve;
                }
            }
            bufRead.clear();
        }
        fis.close();
        return getElapsedTimeMillis()-lStart;
    }

    private long performWriteTest(File f,ByteBuffer buf,final long lWriteCount,Random prng) 
    throws IOException{
        long lStart=getElapsedTimeMillis();
        FileOutputStream fos=new FileOutputStream(f);
        FileChannel fc=fos.getChannel();
        for (long lBlock=0; !isPendingInterrupt() &&  (lBlock<lWriteCount); lBlock++) {
            if (prng!=null) {
                BlockDataGenerator.fillBufferRandom(buf,prng);
            }
            fc.write(buf);
            buf.rewind();
        }
        fos.close();
        return getElapsedTimeMillis()-lStart;
    }
    private void corruptFile(File f,final long lFileSize,final int iBlkSize) throws IOException{
        final long iOffset=(long)(Math.random()*lFileSize);
        RandomAccessFile raf=new RandomAccessFile(f,"rw");
        raf.seek(iOffset);
        byte b=raf.readByte();
        System.err.println("Corrupting File ["+f.getName()+"] at offset "+iOffset+
                           " block ("+(iOffset/iBlkSize)+") "+
                           "changing "+Byte.toString(b)+" to "+Byte.toString((byte)(b^0x7f)));
        raf.seek(iOffset);
        raf.writeByte(b^0x7f);
        raf.close();
    }

    /**
     * Run the actual test. Implementations must call preRun(),
     * and postRun() at the appropriate points in their method.
     * <BR><BR>
     * Additionally, implementations should check the
     * isPendingInterrupt() method to see if they should terminate
     * processing because of an interrupt.
     */
    public void run() {
        Throwable tError=null;
        lSeed=System.currentTimeMillis();
        Random prng=new Random(lSeed);
        final int iBlockSize=(int)TestUtil.getPowerOf2Value(getProperty("BlockSize","16k"));
        final long lFileSize=TestUtil.getPowerOf2Value(getProperty("FileSize","1G"));
        final boolean bVerify="true".equalsIgnoreCase(getProperty("VerifyRead","true"));
        final long lWrites=lFileSize/iBlockSize;

        ByteBuffer buf=ByteBuffer.allocateDirect(iBlockSize);

        File f=new File(getProperty("TestPath",".")+File.separator+Thread.currentThread().getName()+"-SequentialReadWrite.tmp");
        try {

            if (f.exists() && !f.delete()) {
                throw new IOException("Error Deleting file "+f);
            }
            NumberFormat nf=NumberFormat.getInstance();
            preRun();
            long lWrite=performWriteTest(f,buf,lWrites,prng);

            if (testCorruption()) {
                corruptFile(f,lFileSize,iBlockSize);
            }

            prng=new Random(lSeed);

            long lRead=performReadTest(f,buf,lWrites,bVerify,prng);

            if (!f.delete()) {
                throw new IOException("Error deleting file "+f);
            }
            double dValue=((double)(lFileSize>>20))/
                          ((double)lWrite/(double)1000);

            TestResult tr=new TestResult("Sequential Write Throughput",nf.format(dValue),"MB/Sec");
            addResult(tr);

            dValue=((double)(lFileSize>>20))/
                   ((double)lRead/(double)1000);

            tr=new TestResult("Sequential Read Throughput",nf.format(dValue),"MB/Sec");
            addResult(tr);
        } catch (IOException ioe) {
            tError=ioe;
            reportException(ioe);
        }
        if (tError!=null) {
            DataVerificationError dve=new DataVerificationError("SequentialIOTest.run()",tError);
            this.reportException(dve);
            throw dve;
        }
        postRun();
    }
}
