package com.mhsoftware.jonnie;
/*
    MH Software, Inc.
    5023 W 120th Ave, #311
    Broomfield, CO 80020

    Voice: 303 438 9585
    Web: http://www.mhsoftware.com/

    Copyright 2008, MH Software, Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.w3c.dom.*;

/** 
 * This class is the actual class that will execute the tests. 
 * 
 * @author George Sexton
 */
public class TestRunner implements Runnable,Cloneable {

    private List<JonnieTest> lstTests;
    private boolean pendingInterrupt=false;
    private long iterationCount=1;
	private long totalIterations;
    private TestEventListener tel;

    /**
     * When an object implementing interface <code>Runnable</code> is used
     * to create a thread, starting the thread causes the object's
     * <code>run</code> method to be called in that separately executing
     * thread.
     * <p>
     * The general contract of the method <code>run</code> is that it may
     * take any action whatsoever.
     *
     * @see     java.lang.Thread#run()
     */
    public void run() {
        while (!pendingInterrupt && (iterationCount>0)) {
			tel.log("Beginning test cycle "+(totalIterations-iterationCount+1)+" of "+totalIterations+".");
            for (JonnieTest jt : lstTests) {
                if (pendingInterrupt) {
                    break;
                } else {
                    jt.run();
                }
            }
            iterationCount--;
        }
    }

    public Object clone() throws CloneNotSupportedException {
        // We want to do cloning, but we need to make sure that the test 
        // objects are clones, and not references to the same objects as 
        // another  thread
        TestRunner tr=(TestRunner)super.clone();
        tr.clearTests();
        for (JonnieTest jt : lstTests) {
            tr.addTest((JonnieTest)jt.clone());
        }
        return tr;
    }

    public List<JonnieTest> getTests(){
        return lstTests;
    }

    public void clearTests(){
        lstTests=null;
    }

    /** 
     * Add a test to be run by this instance.
     * 
     * @param tst
     */
    public void addTest(JonnieTest tst){
        if (lstTests==null) {
            lstTests=new ArrayList<JonnieTest>();
        }
        lstTests.add(tst);
    }

    public void interrupt(){
        pendingInterrupt=true;
        if (lstTests!=null) {
            for (JonnieTest jt : lstTests) {
                if (jt.isRunning()) {
                    jt.interrupt();
                }
            }
        }
    }

    /** 
     * Set the number of iterations this thread should execute.  
     * 
     * @param lValue If lValue < 0, then Long.MAX_VALUE is used.
     */
    public void setIterationCount(long lValue){
        if (lValue<0) {
            iterationCount=Long.MAX_VALUE;
        } else {
            iterationCount=lValue;
        }
		totalIterations=iterationCount;
    }

    private static Properties getProperties(Node doc){
        Properties prop=new Properties();
        Node[] propList=XMLBase.getElementNodes(doc,"Property");
        for (int i=0; i < propList.length; i++) {
            String PropName=XMLBase.getElementText(propList[i],"Name");
            String PropValue=XMLBase.getElementText(propList[i],"Value");
            prop.setProperty(PropName,PropValue);
        }
        return prop;
    }

    public static void usage(){
		System.out.println("\nJonnie - Java Disk IO Test Program\n");
		System.out.println("Copyright (C) 2008 MH Software, Inc.");
		System.out.println("This program comes with ABSOLUTELY NO WARRANTY;\n"+
						   "This is free software, and you are welcome to redistribute it\n"+
						   "under certain conditions; refer to the file LICENSE for complete\n"+
						   "details.");

        System.out.println("\n\nUsage: java -server -jar jonnie.jar tests.xml\n\n");
    }

    public void setListener(TestEventListener lstnr){
	tel=lstnr;
	}

    public TestEventListener getListener() {
	return tel;
	}

    public static void main(String[] argv) 
    throws InterruptedException, 
    CloneNotSupportedException,
    ClassNotFoundException,
    IllegalAccessException,
    InstantiationException {
        if (argv.length!=1) {
            usage();
            return;

        }
        TestRunner tr=new TestRunner();
	tr.setListener(new TestEventListenerImpl());
        Node doc=XMLBase.parseXMLDocument(XMLBase.readXMLFile(argv[0]));
        Properties prop=getProperties(doc);
        Node[] testList=XMLBase.getElementNodes(doc,"Test");
        final int iterationCount=Integer.parseInt(prop.getProperty("IterationCount","1"));
        tr.setIterationCount(iterationCount);

        for (int iTest=0; iTest < testList.length; iTest++) {
            String cName=XMLBase.getElementText(testList[iTest],"ClassName");
            Class cl=Class.forName(cName);
            JonnieTest jt=(JonnieTest)cl.newInstance();
            jt.addProperties(prop);
            jt.addProperties(getProperties(testList[iTest]));
            jt.addEventListener(tr.getListener());
            tr.addTest(jt);
        }

        final int threadCount=Integer.parseInt(prop.getProperty("ThreadCount","1"));
        Thread th;
        if (threadCount>1) {
            ThreadGroup tg=new ThreadGroup("TestRunner");
            for (int i=0; i < threadCount; i++ ) {
                TestRunner trClone=(TestRunner)tr.clone();
                th=new Thread(tg,trClone,"Thread-"+i);
                System.out.println("Starting Thread - "+i);
                th.start();
            }
            Thread.currentThread().sleep(1000);
            while (tg.activeCount()>0) {
                Thread.currentThread().sleep(500);
                if (tr.getListener().wasExceptionThrown()) {
                    // If any thread throws an exception, we want to terminate the testing
                    System.exit(1);
                }
            }
        } else {
            th=new Thread(tr);
            th.start();
            th.join();
        }
    }
}
